//+------------------------------------------------------------------+
//|                                              Struct_OpenCL_3.mq5 |
//|                                              Copyright 2012, MIG |
//|                                                    Miguzov@bk.ru |
//+------------------------------------------------------------------+
#property copyright "Copyright 2012, MIG"
#property link      "Miguzov@bk.ru"
#property version   "1.03"

input int Size_Test = 2000;  // 

//    OpenCL
struct OpenCLRates                                                                                                                 
{                                                                                                            
   ulong   time;         //                                                                
   float   open;         //                                                                      
   float   high;         //                                                             
   float   low;          //                                                            
   float   close;        //                                                                      
   long    tick_volume;  //                                                                 
   int     spread;       //                                                                             
   char    reserved[4];                                                                        
};  

//  
const string cl_src =
"typedef struct                                                        \r\n"                    
"{                                                                     \r\n"
"   ulong   time;         //                         \r\n"
"   float   open;         //                               \r\n"
"   float   high;         //                      \r\n"
"   float   low;          //                     \r\n"
"   float   close;        //                               \r\n"
"   long    tick_volume;  //   uint                        \r\n"
"   int     spread;       //                                      \r\n"
"   char    reserved[4];                                               \r\n"
"} OpenCLRates;                                                        \r\n"
"__kernel void Work (global OpenCLRates *Rates,                        \r\n" // 
"                    global float *Out)                                \r\n" //  
"{                                                                     \r\n"
"int  Bars = get_global_id(0);                                         \r\n"
"     Out[Bars]=Rates[Bars].close;                                     \r\n"   
"}                                                                     \r\n";
//+------------------------------------------------------------------+
//| Expert initialization function                                   |
//+------------------------------------------------------------------+
int OnInit()
  { 
MqlRates Price[];
OpenCLRates Price_CL[];
float Result_CL[];

ArrayResize(Price,Size_Test);
ArrayResize(Price_CL,Size_Test);
ArrayResize(Result_CL,Size_Test);

ArraySetAsSeries(Price,true);
ArraySetAsSeries(Price_CL,true);
ArraySetAsSeries(Result_CL,true);

ArrayInitialize(Result_CL,0.0);

if (CopyRates(_Symbol,_Period,0,Size_Test,Price)<Size_Test) Print ("   !");

//    
for(int i=0; i<Size_Test; i++) 
     {
     Price_CL[i].time =       (ulong)Price[i].time;
     Price_CL[i].tick_volume = (long)Price[i].tick_volume;
     Price_CL[i].spread =       (int)Price[i].spread;
     Price_CL[i].open =       (float)Price[i].open;
     Price_CL[i].close =      (float)Price[i].close;
     Price_CL[i].high =       (float)Price[i].high;
     Price_CL[i].low =        (float)Price[i].low;
     }      
    
//  OpenCL    
int cl_ctx; //  
int cl_prg; //  
int cl_krn; //  
  
uint  offset[1]={0};
uint  work  [1];
work[0]=(uint)Size_Test; //  

//  OpenCL   
if((cl_ctx=CLContextCreate(0))==-1)
  {
    Print("OpenCL not found! Err: ", GetLastError());
    return(-1);
  }
  
//  OpenCL    
if((cl_prg=CLProgramCreate(cl_ctx,cl_src))==-1)
  {
    CLContextFree(cl_ctx);
    Print("OpenCL program create failed");
    switch(GetLastError())
        {
         case ERR_OPENCL_INVALID_HANDLE:
            Print("Err =    program OpenCL");
            break;
         case ERR_INVALID_PARAMETER:
            Print("Err =   ");
            break;
         case ERR_OPENCL_TOO_LONG_KERNEL_NAME:
            Print("Err =     127 ");
            break;
         case ERR_OPENCL_KERNEL_CREATE:
            Print("Err =      OpenCL.");
            break;
         default: Print("Err =  .");
        }
    return(-1);
  }
  
//     OpenCL  
if((cl_krn=CLKernelCreate(cl_prg,"Work"))==-1)
  {
    CLProgramFree(cl_prg);
    CLContextFree(cl_ctx);
    Print("OpenCL kernel create failed");
    return(-1);
  }
  
//  OpenCL (2 .)
int cl_mem_0 = CLBufferCreate(cl_ctx,Size_Test*sizeof(OpenCLRates),CL_MEM_READ_WRITE); 
int cl_mem_1 = CLBufferCreate(cl_ctx,Size_Test*sizeof(float),CL_MEM_READ_WRITE); 

if((cl_mem_0)==-1||(cl_mem_1)==-1)
  {
    CLKernelFree(cl_krn);
    CLProgramFree(cl_prg);
    CLContextFree(cl_ctx);
    Print("OpenCL buffer create failed");
    switch(GetLastError())
        {
         case ERR_OPENCL_INVALID_HANDLE:
            Print("Err =     OpenCL");
            break;
         case ERR_NOT_ENOUGH_MEMORY:
            Print("Err =  ");
            break;
         case ERR_OPENCL_BUFFER_CREATE:
            Print("Err =    ");
            break;
         default: Print("Err =  .");
        }   
    return(-1);
  }

//  OpenCL     OpenCL
CLSetKernelArgMem(cl_krn,0,cl_mem_0);
CLSetKernelArgMem(cl_krn,1,cl_mem_1);
  
//    OpenCL
if (!CLBufferWrite(cl_mem_0,Price_CL))
{
CLBufferFree(cl_mem_0);
CLBufferFree(cl_mem_1);
CLKernelFree(cl_krn);
CLProgramFree(cl_prg);
CLContextFree(cl_ctx);
Print ("Err    !");
}
 
//C  GPU
if (!CLExecute(cl_krn,1,offset,work)) Print ("Err    OpenCL!");

//   
if (!CLBufferRead(cl_mem_1,Result_CL)) Print ("Err   Result_CL!");
 
//
Print(" Close   110: ",Price[110].close);
Print(" Close   110,  OpenCL: ",Result_CL[110]);

// 
CLBufferFree(cl_mem_0);
CLBufferFree(cl_mem_1);
CLKernelFree(cl_krn);
CLProgramFree(cl_prg);
CLContextFree(cl_ctx);
                                                         
return(0);
}

  